const FRAMEDBG = false;
const RACEDBG = false;

class PcCaptureSessionFrame
{
  async loadImage(image, url) {
    return new Promise((resolve, reject) => {
      image.onload = () => resolve(image);
      image.onerror = reject;
      image.src = url;
    });
  }

  async _prepareFrame(data, canvas, dataIsCanvas)
  {
    PcConfig.dbg(CRASHDBG|FRAMEDBG, "entering with dataIsCanvas", dataIsCanvas);
    if (dataIsCanvas)
    {
      let context = canvas.getContext('2d');
      canvas.width = data.width;
      canvas.height = data.height;
      context.drawImage(data, 0, 0, data.width, data.height);

      if (PcConfig.scenario.cameraOrientation=="portrait" || PcConfig.scenario.cameraOrientation=="portraitInverted")
      {
        let destW = canvas.width;
        let destH = canvas.height;

        let newcanvas = this.cloneCanvas(canvas);

        context.clearRect(0,0,canvas.width,canvas.height);
        canvas.width = destH;
        canvas.height = destW;
        context.save();
        context.translate(canvas.width/2, canvas.height/2);

        if (PcConfig.scenario.cameraOrientation=="portrait")
        {
          context.rotate(-90*Math.PI/180);
        }
        else {
          context.rotate(90*Math.PI/180);
        }

        context.drawImage(newcanvas, -destW/2, -destH/2, destW, destH);
        context.restore();
      }
      
    }
    else { // it is a file
      this.imageURL = data;
    
      PcConfig.dbg(CRASHDBG|FRAMEDBG, "get image from url", this.imageURL);
      let image = document.createElement('img');

      PcConfig.dbg(RACEDBG, "***WAIT FOR LOADIMAGE", this.imageURL);

      /* then() ensures that the promise is settled, the first argument is the callback of success/fullfil settlement, the second argument is the callback of failure (none here) */

      /*await this.loadImage(image, this.imageURL) //we wait for the promise: 
      .then( (image) => { // promise fulfill callback :and then we collect the results 
          PcConfig.dbg(RACEDBG, "***CALL IMAGELOADED", this.imageURL);
          this.imageLoaded(canvas, image);
        }
      );
      */

      // other way to do it (unclear how to get the failure settlement):
      
      image = await this.loadImage(image, this.imageURL);
      PcConfig.dbg(RACEDBG, "***CALL IMAGELOADED", this.imageURL);
      this.imageLoaded(canvas, image);
      
      /* 
        WARNING: the following legacy code would give us a race condition where imageLoaded was called *AFTER* rest of processing (composition etc) 
        => we would end up with a default black canvas of 300x150 instead of the picture with transparent background
      */
      /*
      image.onload = () => this.imageLoaded(canvas, image); //  THIS IS ASYNCHRONOUS
      image.src = data; // pure js: actually loads the image and calls onload afterwards 
      */
    }

    PcConfig.dbg(CRASHDBG|FRAMEDBG, "exiting");

  }

  constructor() // constructors cannot be async, use defensive programming with a new api call: initFrame() where all the async stuff is done
  {
    this.canvas = document.createElement('canvas');
    this.filtered = {}; //dict that will contain one frame per applied filter
    this.maxSourceWidth = 10000;
  }

  async initFrame(data, dataIsCanvas) // do / wait for the async stuff
  {
    PcConfig.dbg(FRAMEDBG, "dataIsCanvas:", dataIsCanvas, "data", data);
    await this._prepareFrame(data, this.canvas, dataIsCanvas);
  }

  imageLoaded(canvas, image)
  {
    PcConfig.dbg(CRASHDBG|FRAMEDBG, "entering");
    var w = image.naturalWidth;
    var h = image.naturalHeight;
    var destW, destH;

    if (w>this.maxSourceWidth)
    {
      destW = this.maxSourceWidth;
      destH = destW * h/w;
    }
    else {
      destW = w;
      destH = h;
    }
    let context = canvas.getContext('2d');

    canvas.width = destW;
    canvas.height = destH;
    context.drawImage(image, 0, 0, destW, destH);

    if (PcConfig.scenario.cameraOrientation=="portrait" || PcConfig.scenario.cameraOrientation=="portraitInverted")
    {
      let newcanvas = this.cloneCanvas(canvas);

      context.clearRect(0,0,canvas.width,canvas.height);
      canvas.width = destH;
      canvas.height = destW;
      context.save();
      context.translate(canvas.width/2, canvas.height/2);

      if (PcConfig.scenario.cameraOrientation=="portrait")
      {
        context.rotate(-90*Math.PI/180);
      }
      else {
        context.rotate(90*Math.PI/180);
      }

      context.drawImage(newcanvas, -destW/2, -destH/2, destW, destH);
      context.restore();
    }

    PcConfig.dbg(CRASHDBG|FRAMEDBG, "exiting", image.src, "(check if I exited after _prepareFrame() )");
  }

  filter(filter_dict, index)
  {
    let filtered;
    let canvas;
    let mode;
    // returns a filtered canvas from PcFilter switch-case filtering code depending on filter name
    
    mode = "normal";
    canvas = this.canvas;
    filtered = this.filtered;
    
    if (filtered.hasOwnProperty(filter_dict.id)) { // filter dict already contins the filtered picture -> don't do it anymore
      return;
    }

    let fd;
    if (filter_dict.id) {
      fd = filter_dict.filterName.toString(); //id is a combination of filtername and args with lots of spaces in between and ":" separator before that
    } else { 
      fd = "undefined";
    }
    var filter = new PcFilter(filter_dict);
    filtered[filter_dict.id] = filter.apply(canvas, index);
  }

  /* per Frame framing now takes predef_frame_landmarks.
  it is used to potentially avoid recomputing landmarks 
  */
  async framing(framing, predef_frame_landmarks, faceRatio, centerRatio, frameSize, alert)
  {
    PcConfig.dbg(FRAMINGDBG, "PCSF: framing: predef_frame_landmarks", JSON.stringify(predef_frame_landmarks));
    PcConfig.dbg(FRAMINGDBG, "PCSF: framing: faceRatio",  faceRatio, "centerRatio", JSON.stringify(centerRatio), "frameSize", JSON.stringify(frameSize), "au mieux",  !alert);
    let frame_landmarks;
    [ this.canvas, frame_landmarks ]   = await framing.apply(
      predef_frame_landmarks, 
      this.canvas,
      faceRatio,
      centerRatio,
      frameSize,
      alert); // framing is a class instance from PcCaptureSession, yuk

      PcConfig.dbg(FRAMINGDBG, "PCSF: framing -> frame_landmarks =", JSON.stringify(frame_landmarks));

      // return the Frame landmarks
      return frame_landmarks;
  }

  cloneCanvas(oldCanvas)
  {
    var newCanvas = document.createElement('canvas');
    var context = newCanvas.getContext('2d');

    newCanvas.width = oldCanvas.width;
    newCanvas.height = oldCanvas.height;

    context.drawImage(oldCanvas, 0, 0);

    return newCanvas;
  }
  
  delete()
  {
    for (var c of [this.canvas]) { // "var. of" works like "x in"  in python .. while "var. in" gives index of an array  
      c = null;
    }
    PcConfig.dbg(FRAMEDBG, "filtered:", JSON.stringify(this.filtered));
    for (var filtered of [this.filtered]) 
    {
      for (var p in filtered) { // filtered is a dict, var..in is required to iterate on keys
       if (filtered.hasOwnProperty(p))
       {
          filtered[p] = null;
        }
      }
      filtered = null;
    }

  }
}
